C
      SUBROUTINE UMAT(STRESS,STATEV,DDSDDE,SSE,SPD,SCD,
     1 RPL,DDSDDT,DRPLDE,DRPLDT,
     2 STRAN,DSTRAN,TIME,DTIME,TEMP,DTEMP,PREDEF,DPRED,CMNAME,
     3 NDI,NSHR,NTENS,NSTATEV,PROPS,NPROPS,COORDS,DROT,PNEWDT,
     4 CELENT,DFGRD0,DFGRD1,NOEL,NPT,LAYER,KSPT,KSTEP,KINC)
C
      INCLUDE 'ABA_PARAM.INC'
C
      CHARACTER*8 CMNAME
      DIMENSION STRESS(NTENS),STATEV(NSTATEV),
     1 DDSDDE(NTENS,NTENS),DDSDDT(NTENS),DRPLDE(NTENS),
     2 STRAN(NTENS),DSTRAN(NTENS),TIME(2),PREDEF(1),DPRED(1),
     3 PROPS(NPROPS),COORDS(3),DROT(3,3),DFGRD0(3,3),DFGRD1(3,3)
C LOCAL ARRAYS
C ----------------------------------------------------------------
C EELAS - LOGARITHMIC ELASTIC STRAINS
C EELASP - PRINCIPAL ELASTIC STRAINS
C BBAR - DEVIATORIC RIGHT CAUCHY-GREEN TENSOR
C BBARP - PRINCIPAL VALUES OF BBAR
C BBARN - PRINCIPAL DIRECTION OF BBAR (AND EELAS)
C DISTGR - DEVIATORIC DEFORMATION GRADIENT (DISTORTION TENSOR)
C BBTOT - DEVIATORIC LEFT CAUCHY-GREEN STRAIN
C BH - THE 4-TENSOR H ON MATRIX (VOIGT) FORM
CCC DASHPOT 1
C Ci - INELASTIC RIGHT CAUCHY-GREEN STRAIN
C CiINV - INVERSE(Ci), STATE-VARIABLE
C Be - ELASTIC LEFT CAUCHY-GREEN STRAIN
C TAUNEQ - NEQ KIRCHHOFF STRESS
C sigVISC - NEQ (VISCOUS) CAUCHY STRESS
C TAUVEC - PRINCPIAL VALUES OF THE NEQ KIRCHHOFF STRESS
C DLTR - PRINCIPAL VALUES OF Be_TRIAL
C DPVEC - PRINCIPAL DIRECTIONS OF Be AND TAU_NEQ
C CALG - ALGORITHMIC TANGENT MODULUS
C DDSDDE2 - NEQ MATERIAL TANGENT ON MATRIX (VOIGT) FORM
CCC DASHPOT 2
C Ci2 - INELASTIC RIGHT CAUCHY-GREEN STRAIN
C CiINV - INVERSE(Ci2), STATE-VARIABLE
C Be2 - ELASTIC LEFT CAUCHY-GREEN STRAIN
C TAUNEQ2 - NEQ KIRCHHOFF STRESS
C sigVISC2 - NEQ (VISCOUS) CAUCHY STRESS
C TAUVEC2 - PRINCPIAL VALUES OF THE NEQ KIRCHHOFF STRESS
C DLTR2 - PRINCIPAL VALUES OF Be_TRIAL
C DPVEC2 - PRINCIPAL DIRECTIONS OF Be AND TAU_NEQ
C CALG2 - ALGORITHMIC TANGENT MODULUS
C DDSDDE22 - NEQ MATERIAL TANGENT ON MATRIX (VOIGT) FORM
C ----------------------------------------------------------------
C
      DIMENSION EELAS(6), EELASP(3), BBAR(6), BBARP(3), BBARN(3, 3),
     1 DISTGR(3,3), BBTOT(3,3), CiINV0(3,3), CiINV1(3,3), Be(3,3),
     2 TAUNEQ(3,3), TAUVEC(3), DLTR(3), DPVEC(3,3), CALG(3,3),
     3 DDSDDE2(6,6), BeINV(3,3), CiTEMP(3,3), Ci(3,3), sigVISC(6),
     4 CiINV02(3,3), CiINV12(3,3), Be2(3,3),
     5 TAUNEQ2(3,3), TAUVEC2(3), DLTR2(3), DPVEC2(3,3), CALG2(3,3),
     6 DDSDDE22(6,6), BeINV2(3,3), CiTEMP2(3,3), Ci2(3,3), sigVISC2(6)
C
      PARAMETER(ZERO=0.D0, ONE=1.D0, TWO=2.D0, THREE=3.D0, FOUR=4.D0,
     1 SIX=6.D0)
C
C ----------------------------------------------------------------
C UMAT FOR COMPRESSIBLE NEO-HOOKEAN HYPERELASTICITY
C WITH FV-THEORY AS DESCRIBED BY REESE & GOVINDJEE
C ----------------------------------------------------------------
C PROPS(1) - MU
C PROPS(2) - K
C PROPS(3) - ETADEV
C PROPS(4) - ETAVOL
C PROPS(5) - BETA
C ----------------------------------------------------------------
C
C ELASTIC PROPERTIES
C
      C10=PROPS(1)
      D1=PROPS(2)
      VISCdev=PROPS(3)
      VISCvol=PROPS(4)
      BETA=PROPS(5)
      VISCdev2=PROPS(6)
      VISCvol2=PROPS(7)
      BETA2=PROPS(8)
C      
      DMU=C10*TWO
      DK=TWO/D1
C
C JACOBIAN AND DISTORTION TENSOR
C
      DET=DFGRD1(1, 1)*DFGRD1(2, 2)*DFGRD1(3, 3)
     1 -DFGRD1(1, 2)*DFGRD1(2, 1)*DFGRD1(3, 3)
      IF(NSHR.EQ.3) THEN
      DET=DET+DFGRD1(1, 2)*DFGRD1(2, 3)*DFGRD1(3, 1)
     1 +DFGRD1(1, 3)*DFGRD1(3, 2)*DFGRD1(2, 1)
     2 -DFGRD1(1, 3)*DFGRD1(3,1)*DFGRD1(2, 2)
     3 -DFGRD1(2, 3)*DFGRD1(3, 2)*DFGRD1(1, 1)
      END IF
      SCALE=DET**(-ONE/THREE)
      DO K1=1, 3
        DO K2=1, 3
        DISTGR(K2, K1)=SCALE*DFGRD1(K2, K1)
        END DO
      END DO
C CALCULATE DEVIATORIC LEFT CAUCHY-GREEN DEFORMATION TENSOR
C
      BBAR(1)=DISTGR(1, 1)**2+DISTGR(1, 2)**2+DISTGR(1, 3)**2
      BBAR(2)=DISTGR(2, 1)**2+DISTGR(2, 2)**2+DISTGR(2, 3)**2
      BBAR(3)=DISTGR(3, 3)**2+DISTGR(3, 1)**2+DISTGR(3, 2)**2
      BBAR(4)=DISTGR(1, 1)*DISTGR(2, 1)+DISTGR(1, 2)*DISTGR(2, 2)
     1 +DISTGR(1, 3)*DISTGR(2, 3)
      IF(NSHR.EQ.3) THEN
      BBAR(5)=DISTGR(1, 1)*DISTGR(3, 1)+DISTGR(1, 2)*DISTGR(3, 2)
     1 +DISTGR(1, 3)*DISTGR(3, 3)
      BBAR(6)=DISTGR(2, 1)*DISTGR(3, 1)+DISTGR(2, 2)*DISTGR(3, 2)
     1 +DISTGR(2, 3)*DISTGR(3, 3)
      END IF
C
      BBTOT(1,1) = BBAR(1)
      BBTOT(1,2) = BBAR(4)
      BBTOT(1,3) = BBAR(5)
      BBTOT(2,1) = BBAR(4)
      BBTOT(2,2) = BBAR(2)
      BBTOT(2,3) = BBAR(6)
      BBTOT(3,1) = BBAR(5)
      BBTOT(3,2) = BBAR(6)
      BBTOT(3,3) = BBAR(3)
C
C     DEFINING THE INITIAL STATE-VARIABLE INV(Ci)
C    
      IF (KSTEP.EQ.1 .AND. KINC.EQ.1) THEN
       STATEV(1)=ONE
       STATEV(2)=ONE
       STATEV(3)=ONE
       STATEV(4)=ZERO
       STATEV(5)=ZERO
       STATEV(6)=ZERO
       STATEV(7)=ONE
       STATEV(8)=ONE
       STATEV(9)=ONE
       STATEV(10)=ZERO
       STATEV(11)=ZERO
       STATEV(12)=ZERO
      END IF
C      
      CiINV0(1,1)=STATEV(1)
      CiINV0(1,2)=STATEV(4)
      CiINV0(1,3)=STATEV(5)
      CiINV0(2,1)=STATEV(4)
      CiINV0(2,2)=STATEV(2)
      CiINV0(2,3)=STATEV(6)
      CiINV0(3,1)=STATEV(5)
      CiINV0(3,2)=STATEV(6)
      CiINV0(3,3)=STATEV(3)
C
      CiINV02(1,1)=STATEV(7)
      CiINV02(1,2)=STATEV(10)
      CiINV02(1,3)=STATEV(11)
      CiINV02(2,1)=STATEV(10)
      CiINV02(2,2)=STATEV(8)
      CiINV02(2,3)=STATEV(12)
      CiINV02(3,1)=STATEV(11)
      CiINV02(3,2)=STATEV(12)
      CiINV02(3,3)=STATEV(9)
C
C     CALCULATION OF ELASTIC LEFT CAUCHY-GREEN STRAIN AND 
C     KIRCHHOFF STRESS FOR THE NON-EQ (VISCOUS) DEFORMATION (1)
C
      DTID=DTIME
        CALL BECALCMOD (DMU, DK, BETA,NDI,NSHR,
     1            VISCdev, VISCvol, DTID,
     2            DFGRD1, CiINV0, Be, TAUNEQ, TAUVEC, DLTR, DPVEC, CALG)
    
C     MATERIAL TANGENT OF THE INELASTIC DEFORMATION
C
        CALL DNEQTANGENT (DET, TAUVEC, DLTR, DPVEC,
     1                        CALG, TAUNEQ, DDSDDE2)
C
C
C     CALCULATION OF ELASTIC LEFT CAUCHY-GREEN STRAIN AND 
C     KIRCHHOFF STRESS FOR THE NON-EQ (VISCOUS) DEFORMATION (2)
C
        CALL BECALCMOD (DMU, DK, BETA2,NDI,NSHR,
     1            VISCdev2, VISCvol2, DTID,
     2            DFGRD1, CiINV02, Be2, TAUNEQ2, 
     3            TAUVEC2, DLTR2, DPVEC2, CALG2)
    
C     MATERIAL TANGENT OF THE INELASTIC DEFORMATION
C
        CALL DNEQTANGENT (DET, TAUVEC2, DLTR2, DPVEC2,
     1                        CALG2, TAUNEQ2, DDSDDE22)
C
C     CALCULATION OF RIGHT CAUCHY-GREEN STRAIN OF THE NON-EQ
C     DEFORMATION THE INVERSE TO UPDATE THE STATE-VARIABLES
C
      CALL M33INV(Be, BeINV, FLAG)
      CiTEMP=MATMUL(BeINV,DFGRD1)
      Ci=MATMUL(TRANSPOSE(DFGRD1),CiTEMP)
      CALL M33INV(Ci, CiINV1, FLAG)
C
      CALL M33INV(Be2, BeINV2, FLAG)
      CiTEMP2=MATMUL(BeINV2,DFGRD1)
      Ci2=MATMUL(TRANSPOSE(DFGRD1),CiTEMP2)
      CALL M33INV(Ci2, CiINV12, FLAG)
C     
      STATEV(1)=CiINV1(1,1)
      STATEV(2)=CiINV1(2,2)
      STATEV(3)=CiINV1(3,3)
      STATEV(4)=CiINV1(1,2)
      STATEV(5)=CiINV1(1,3)
      STATEV(6)=CiINV1(2,3)
C
C     
      STATEV(7)=CiINV12(1,1)
      STATEV(8)=CiINV12(2,2)
      STATEV(9)=CiINV12(3,3)
      STATEV(10)=CiINV12(1,2)
      STATEV(11)=CiINV12(1,3)
      STATEV(12)=CiINV12(2,3)
C
C     CALCUATION OF NON-EQ STRESS (VISCOUS STRESS) (1)
C
      sigVISC(1)=TAUNEQ(1,1)/DET
      sigVISC(2)=TAUNEQ(2,2)/DET
      sigVISC(3)=TAUNEQ(3,3)/DET
      sigVISC(4)=TAUNEQ(1,2)/DET
      sigVISC(5)=TAUNEQ(1,3)/DET
      sigVISC(6)=TAUNEQ(2,3)/DET
C
C     CALCUATION OF NON-EQ STRESS (VISCOUS STRESS) (2)
C
      sigVISC2(1)=TAUNEQ2(1,1)/DET
      sigVISC2(2)=TAUNEQ2(2,2)/DET
      sigVISC2(3)=TAUNEQ2(3,3)/DET
      sigVISC2(4)=TAUNEQ2(1,2)/DET
      sigVISC2(5)=TAUNEQ2(1,3)/DET
      sigVISC2(6)=TAUNEQ2(2,3)/DET
C
C CALCULATE THE STRESS
C
      TRBBAR=(BBAR(1)+BBAR(2)+BBAR(3))/THREE
      EG=TWO*C10/DET
      EK=TWO/D1*(TWO*DET-ONE)
      PR=TWO/D1*(DET-ONE)
      DO K1=1,NDI
        STRESS(K1)=EG*(BBAR(K1)-TRBBAR)+PR+sigVISC(K1)+sigVISC2(K1)
      END DO
      DO K1=NDI+1,NDI+NSHR
        STRESS(K1)=EG*BBAR(K1)+sigVISC(K1)+sigVISC2(K1)
      END DO
C CALCULATE THE STIFFNESS
C
      EG23=EG*TWO/THREE
      DDSDDE(1, 1)= EG23*(BBAR(1)+TRBBAR)+EK+DDSDDE2(1,1)+DDSDDE22(1,1)
      DDSDDE(2, 2)= EG23*(BBAR(2)+TRBBAR)+EK+DDSDDE2(2,2)+DDSDDE22(2,2)
      DDSDDE(3, 3)= EG23*(BBAR(3)+TRBBAR)+EK+DDSDDE2(3,3)+DDSDDE22(3,3)
      DDSDDE(1, 2)=-EG23*(BBAR(1)+BBAR(2)-TRBBAR)+EK+DDSDDE2(1, 2)
     1             +DDSDDE22(1, 2)
      DDSDDE(1, 3)=-EG23*(BBAR(1)+BBAR(3)-TRBBAR)+EK+DDSDDE2(1, 3)
     1             +DDSDDE22(1, 3)
      DDSDDE(2, 3)=-EG23*(BBAR(2)+BBAR(3)-TRBBAR)+EK+DDSDDE2(2, 3)
     1             +DDSDDE22(2, 3)
      DDSDDE(1, 4)= EG23*BBAR(4)/TWO+DDSDDE2(1, 4)+DDSDDE22(1, 4)
      DDSDDE(2, 4)= EG23*BBAR(4)/TWO+DDSDDE2(2, 4)+DDSDDE22(2, 4)
      DDSDDE(3, 4)=-EG23*BBAR(4)+DDSDDE2(3, 4)+DDSDDE22(3, 4)
      DDSDDE(4, 4)= EG*(BBAR(1)+BBAR(2))/TWO+DDSDDE2(4,4)+DDSDDE22(4,4)
      IF(NSHR.EQ.3) THEN
        DDSDDE(1, 5)= EG23*BBAR(5)/TWO+DDSDDE2(1,5)+DDSDDE22(1,5)
        DDSDDE(2, 5)=-EG23*BBAR(5)+DDSDDE2(2,5)+DDSDDE22(2,5)
        DDSDDE(3, 5)= EG23*BBAR(5)/TWO+DDSDDE2(3,5)+DDSDDE22(3,5)
        DDSDDE(1, 6)=-EG23*BBAR(6)+DDSDDE2(1,6)+DDSDDE22(1,6)
        DDSDDE(2, 6)= EG23*BBAR(6)/TWO+DDSDDE2(2,6)+DDSDDE22(2,6)
        DDSDDE(3, 6)= EG23*BBAR(6)/TWO+DDSDDE2(3,6)+DDSDDE22(3,6)
        DDSDDE(5, 5)= EG*(BBAR(1)+BBAR(3))/TWO+DDSDDE2(5,5)
     1               +DDSDDE22(5,5)
        DDSDDE(6, 6)= EG*(BBAR(2)+BBAR(3))/TWO+DDSDDE2(6,6)
     1               +DDSDDE22(6,6)
        DDSDDE(4,5)= EG*BBAR(6)/TWO+DDSDDE2(4,5)+DDSDDE22(4,5)
        DDSDDE(4,6)= EG*BBAR(5)/TWO+DDSDDE2(4,6)+DDSDDE22(4,6)
        DDSDDE(5,6)= EG*BBAR(4)/TWO+DDSDDE2(5,6)+DDSDDE22(5,6)
      END IF
      DO K1=1, NTENS
        DO K2=1, K1-1
        DDSDDE(K1, K2)=DDSDDE(K2, K1)
        END DO
      END DO
C     CALCULATE SOME OUTPUT VARIABLES
C
      STATEV(13)=DET
      STATEV(14)=STRESS(2)
      STATEV(15)=STRESS(2)-sigVISC(2)-sigVISC2(2)
      STATEV(16)=sigVISC(2)
      STATEV(17)=sigVISC2(2)
      STATEV(18)=DFGRD1(2,2)
      STATEV(19)=DFGRD1(1,1)
      STATEV(20)=Be(2,2)
      STATEV(21)=Be2(2,2)
      STATEV(22)=Be(1,1)
      STATEV(23)=Be2(1,1)
      STATEV(24)=-LOG(DFGRD1(1,1))/LOG(DFGRD1(2,2))
C
C
      RETURN
      END
      
!***********************************************************************************************************************************
!  BECALCMOD -  CALCULATE ELASTIC LEFT CAUCHY STRAIN AND NEQ-KIRCHOFF STRESS
!  
!  INPUT VARIABLES:
!  DFGRD = DEFORMATION GRADIENT OF STEP K
!  CiINV = INVERSE INELASTIC (VISCOUS) RIGHT CAUCHY-GREEN STRAIN TENSOR AT STEP K-1
!  TAU   = KIRCHOFF STRESS AT STEP K+1
!  MATERIALPARAMETERS = DMU, DK, BETA, VISCdev, VISCvol
!  
!  OUTPUT VARIABLES:
!  Be = ELASTIC LEFT CAUCHY STRAIN
!  TAUNEQ = NEQ-KIRCHOFF STRESS
!***********************************************************************************************************************************
C
      SUBROUTINE BECALCMOD (DMU, DK, BETA, NDI,NSHR,
     1                   VISCdev, VISCvol, DTIME,
     1                   DFGRD, CiINV, 
     1                   Be, TAUNEQ, TAUVEC, DLTR, DPVEC, CALG)

      IMPLICIT NONE

      DOUBLE PRECISION, INTENT(IN) :: DMU
      DOUBLE PRECISION, INTENT(IN) :: DK
      DOUBLE PRECISION, INTENT(IN) :: BETA
      DOUBLE PRECISION, INTENT(IN) :: NDI
      DOUBLE PRECISION, INTENT(IN) :: NSHR
      DOUBLE PRECISION, INTENT(IN) :: VISCdev
      DOUBLE PRECISION, INTENT(IN) :: VISCvol
      DOUBLE PRECISION, INTENT(IN) :: DTIME
C      
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: DFGRD
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: CiINV
C
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: Be
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: TAUNEQ
      DOUBLE PRECISION, DIMENSION(3), INTENT(OUT) :: TAUVEC
      DOUBLE PRECISION, DIMENSION(3), INTENT(OUT) :: DLTR
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: DPVEC
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: CALG
C
      DOUBLE PRECISION, PARAMETER :: TOL = 1.0D-12
C     
      DOUBLE PRECISION, DIMENSION(3,3) :: BeTEMP
      DOUBLE PRECISION, DIMENSION(3,3) :: BeTRIAL
      DOUBLE PRECISION, DIMENSION(6) :: BeTR_
C
      DOUBLE PRECISION, DIMENSION(3) :: PV
      DOUBLE PRECISION, DIMENSION(3,3) :: PD
C
      DOUBLE PRECISION :: R1
      DOUBLE PRECISION :: R2
      DOUBLE PRECISION :: BI1
      DOUBLE PRECISION :: DJE0
      DOUBLE PRECISION, DIMENSION(3) :: Q0
      DOUBLE PRECISION, DIMENSION(3) :: Q1
      DOUBLE PRECISION, DIMENSION(3) :: E0
      DOUBLE PRECISION, DIMENSION(3) :: E1
      DOUBLE PRECISION, DIMENSION(3) :: DE
      DOUBLE PRECISION, DIMENSION(3) :: ETRIAL
C
      DOUBLE PRECISION, DIMENSION(3,3) :: DKMAT
      DOUBLE PRECISION, DIMENSION(3,3) :: DKINV
      DOUBLE PRECISION, DIMENSION(3) :: VR
      DOUBLE PRECISION, DIMENSION(3,3) :: DTDEPS
C
      integer :: K1
      integer :: K2
      integer :: K3
      DOUBLE PRECISION :: DnormDE
      LOGICAL :: FLAG
C
C
C
C      
      BeTEMP = MATMUL(CiINV,TRANSPOSE(DFGRD))
      BeTRIAL = MATMUL(DFGRD,BeTEMP)
      BeTR_(1) = BeTRIAL(1,1)
      BeTR_(2) = BeTRIAL(2,2)
      BeTR_(3) = BeTRIAL(3,3)
      BeTR_(4) = BeTRIAL(1,2)+BeTRIAL(2,1)
      BeTR_(5) = BeTRIAL(1,3)+BeTRIAL(3,1)
      BeTR_(6) = BeTRIAL(2,3)+BeTRIAL(3,2)
C
      CALL SPRIND(BeTR_,PV,PD,2,NDI,NSHR)
C     
      R1=DTIME*BETA*DMU/(2.D0*VISCdev)
      R2=DTIME*BETA*DK/(3.D0*VISCvol)
C     
      Q0(1)=PV(1)
      Q0(2)=PV(2)
      Q0(3)=PV(3)
      ETRIAL(1)=1.D0/2.D0*LOG(PV(1))
      ETRIAL(2)=1.D0/2.D0*LOG(PV(2))
      ETRIAL(3)=1.D0/2.D0*LOG(PV(3))
C
      do K3=1,175,1
        DJE0=(Q0(1)*Q0(2)*Q0(3))**(1.0D0/2.0D0)
        E0(1)=1.D0/2.D0*LOG(Q0(1))
        E0(2)=1.D0/2.D0*LOG(Q0(2))
        E0(3)=1.D0/2.D0*LOG(Q0(3))
C        
        DKMAT(1,1)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(8.0D0*Q0(1)+2.0D0*(Q0(2)+Q0(3)))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0+1.0D0
        DKMAT(1,2)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(1)-4.0D0*Q0(2)+2.0D0*Q0(3))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0
        DKMAT(1,3)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(1)-4.0D0*Q0(3)+2.0D0*Q0(2))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0
        DKMAT(2,1)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(2)-4.0D0*Q0(1)+2.0D0*Q0(3))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0
        DKMAT(2,2)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(8.0D0*Q0(2)+2.0D0*(Q0(1)+Q0(3)))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0+1.0D0
        DKMAT(2,3)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(2)-4.0D0*Q0(3)+2.0D0*Q0(1))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0
        DKMAT(3,1)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(3)-4.0D0*Q0(1)+2.0D0*Q0(2))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0
        DKMAT(3,2)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(3)-4.0D0*Q0(2)+2.0D0*Q0(1))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0
        DKMAT(3,3)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(8.0D0*Q0(3)+2.0D0*(Q0(1)+Q0(2)))
     1                  +R2*(2.0D0*DJE0-1.0D0)*DJE0+1.0D0
C       
        VR(1)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/3.0D0
     1                *(2.0D0*Q0(1)-Q0(2)-Q0(3))
     1                +R2*(DJE0-1.0D0)*DJE0+E0(1)-ETRIAL(1)
        VR(2)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/3.0D0
     1                *(2.0D0*Q0(2)-Q0(1)-Q0(3))
     1                +R2*(DJE0-1.0D0)*DJE0+E0(2)-ETRIAL(2)
        VR(3)=R1*DJE0**(-2.0D0/3.0D0)*1.0D0/3.0D0
     1                *(2.0D0*Q0(3)-Q0(2)-Q0(1))
     1                +R2*(DJE0-1.0D0)*DJE0+E0(3)-ETRIAL(3)
C       
        CALL M33INV(DKMAT, DKINV, FLAG)
        DE(1)=-(DKINV(1,1)*VR(1)+DKINV(1,2)*VR(2)+DKINV(1,3)*VR(3))
        DE(2)=-(DKINV(2,1)*VR(1)+DKINV(2,2)*VR(2)+DKINV(2,3)*VR(3))
        DE(3)=-(DKINV(3,1)*VR(1)+DKINV(3,2)*VR(2)+DKINV(3,3)*VR(3))
C       
        E1(1)=E0(1)+DE(1)
        E1(2)=E0(2)+DE(2)
        E1(3)=E0(3)+DE(3)
C
        DnormDE=(DE(1)**2.D0+DE(2)**2.D0+DE(3)**2.D0)**(1.D0/2.D0)
        if (abs(DnormDE).LT.TOL) then 
            DO K1 = 1, 3
                DO K2 = 1, 3
                    Be(K2,K1)=Q0(1)*PD(1,K2)*PD(1,K1)
     1                       +Q0(2)*PD(2,K2)*PD(2,K1)
     1                       +Q0(3)*PD(3,K2)*PD(3,K1)
                END DO
            END DO
            BI1=Be(1,1)+Be(2,2)+Be(3,3)
            TAUNEQ(1,1)=BETA*DK*(DJE0-1.0D0)*DJE0
     1                 +BETA*DMU*DJE0**(-2.0D0/3.0D0)
     1                 *(Be(1,1)-1.0D0/3.0D0*BI1)
            TAUNEQ(2,2)=BETA*DK*(DJE0-1.0D0)*DJE0
     1                 +BETA*DMU*DJE0**(-2.0D0/3.0D0)
     1                 *(Be(2,2)-1.0D0/3.0D0*BI1)
            TAUNEQ(3,3)=BETA*DK*(DJE0-1.0D0)*DJE0
     1                 +BETA*DMU*DJE0**(-2.0D0/3.0D0)
     1                 *(Be(3,3)-1.0D0/3.0D0*BI1)
            TAUNEQ(1,2)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*Be(1,2)
            TAUNEQ(1,3)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*Be(1,3)
            TAUNEQ(2,3)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*Be(2,3)
            TAUNEQ(2,1)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*Be(2,1)
            TAUNEQ(3,1)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*Be(3,1)
            TAUNEQ(3,2)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*Be(3,2)
            TAUVEC(1)=BETA*DK*(DJE0-1.0D0)*DJE0
     1                 +BETA*DMU*DJE0**(-2.0D0/3.0D0)
     1                 *(Q0(1)-1.0D0/3.0D0*BI1)
            TAUVEC(2)=BETA*DK*(DJE0-1.0D0)*DJE0
     1                 +BETA*DMU*DJE0**(-2.0D0/3.0D0)
     1                 *(Q0(2)-1.0D0/3.0D0*BI1)
            TAUVEC(3)=BETA*DK*(DJE0-1.0D0)*DJE0
     1                 +BETA*DMU*DJE0**(-2.0D0/3.0D0)
     1                 *(Q0(3)-1.0D0/3.0D0*BI1)
            DLTR(1)=EXP(2.0D0*ETRIAL(1))
            DLTR(2)=EXP(2.0D0*ETRIAL(2))
            DLTR(3)=EXP(2.0D0*ETRIAL(3))
            DO K1 = 1, 3
                DO K2 = 1, 3
                    DPVEC(K2,K1)=PD(K2,K1)
                END DO
            END DO
        DTDEPS(1,1)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(8.0D0*Q0(1)+2.0D0*(Q0(2)+Q0(3)))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(1,2)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(1)-4.0D0*Q0(2)+2.0D0*Q0(3))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(1,3)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(1)-4.0D0*Q0(3)+2.0D0*Q0(2))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(2,1)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(2)-4.0D0*Q0(1)+2.0D0*Q0(3))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(2,2)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(8.0D0*Q0(2)+2.0D0*(Q0(1)+Q0(3)))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(2,3)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(2)-4.0D0*Q0(3)+2.0D0*Q0(1))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(3,1)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(3)-4.0D0*Q0(1)+2.0D0*Q0(2))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(3,2)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(-4.0D0*Q0(3)-4.0D0*Q0(2)+2.0D0*Q0(1))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
        DTDEPS(3,3)=BETA*DMU*DJE0**(-2.0D0/3.0D0)*1.0D0/9.0D0
     1                  *(8.0D0*Q0(3)+2.0D0*(Q0(1)+Q0(2)))
     1                  +BETA*DK*(2.0D0*DJE0-1.0D0)*DJE0
            CALG = MATMUL(DTDEPS,DKINV)
            return
        end if 
C     
      Q1(1)=EXP(2.D0*E1(1))
      Q1(2)=EXP(2.D0*E1(2))
      Q1(3)=EXP(2.D0*E1(3))
      Q0(1)=Q1(1)
      Q0(2)=Q1(2)
      Q0(3)=Q1(3)
C        
      END DO
C
      RETURN
C
      END SUBROUTINE BECALCMOD
      

!***********************************************************************************************************************************
!  M33INV  -  Compute the inverse of a 3x3 matrix.
!
!  A       = input 3x3 matrix to be inverted
!  AINV    = output 3x3 inverse of matrix A
!  OK_FLAG = (output) .TRUE. if the input matrix could be inverted, and .FALSE. if the input matrix is singular.
!  Code written by:   David G. Simpson  -   NASA Goddard Space Flight Center
!***********************************************************************************************************************************

      SUBROUTINE M33INV (A, AINV, OK_FLAG)

      IMPLICIT NONE

      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN)  :: A
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: AINV
      LOGICAL, INTENT(OUT) :: OK_FLAG

      DOUBLE PRECISION, PARAMETER :: EPS = 1.0D-16
      DOUBLE PRECISION :: DET
      DOUBLE PRECISION, DIMENSION(3,3) :: COFACTOR


      DET =   A(1,1)*A(2,2)*A(3,3)  
     1       - A(1,1)*A(2,3)*A(3,2)  
     1       - A(1,2)*A(2,1)*A(3,3)  
     1       + A(1,2)*A(2,3)*A(3,1)  
     1       + A(1,3)*A(2,1)*A(3,2)  
     1       - A(1,3)*A(2,2)*A(3,1)

      IF (ABS(DET) .LE. EPS) THEN
         AINV = 0.0D0
         OK_FLAG = .FALSE.
         RETURN
      END IF

      COFACTOR(1,1) = +(A(2,2)*A(3,3)-A(2,3)*A(3,2))
      COFACTOR(1,2) = -(A(2,1)*A(3,3)-A(2,3)*A(3,1))
      COFACTOR(1,3) = +(A(2,1)*A(3,2)-A(2,2)*A(3,1))
      COFACTOR(2,1) = -(A(1,2)*A(3,3)-A(1,3)*A(3,2))
      COFACTOR(2,2) = +(A(1,1)*A(3,3)-A(1,3)*A(3,1))
      COFACTOR(2,3) = -(A(1,1)*A(3,2)-A(1,2)*A(3,1))
      COFACTOR(3,1) = +(A(1,2)*A(2,3)-A(1,3)*A(2,2))
      COFACTOR(3,2) = -(A(1,1)*A(2,3)-A(1,3)*A(2,1))
      COFACTOR(3,3) = +(A(1,1)*A(2,2)-A(1,2)*A(2,1))

      AINV = TRANSPOSE(COFACTOR) / DET

      OK_FLAG = .TRUE.

      RETURN

      END SUBROUTINE M33INV
      

      
      
!***********************************************************************************************************************************
!  DNEQ -  CALCULATE NEQ TANGENT
!  
!  INPUT VARIABLES:
!  DET      =   DETERMINANT OF DEFORMATION GRADIENT
!  TAUVEC   =   VECTOR CONTAINING THE PRINCIPAL KIRCHOFF-STRESSES
!  DLTR     =   VECTOR CONTAINING THE SQUARED PRINCIPAL TRIAL-STRAINS
!  DPVEC    =   MATRIX CONTAINING THE THREE PRINCIPAL DIRECTIONS OF Be
!  CALG     =   ALGORITMIC MODULUS (3,3)
!  TAUNEQ   =   INELASTIC KIRCHOFF STRESS (3,3)
!  
!  OUTPUT VARIABLES:
!  DDSDDE = NEQ MATERIAL-TANGENT (6,6)
!***********************************************************************************************************************************

      SUBROUTINE DNEQTANGENT (DET, TAUVEC, DLTR, DPVEC,
     1                        CALG, TAUNEQ, DDSDDE)
C
      IMPLICIT NONE
C
      DOUBLE PRECISION, INTENT(IN) :: DET
      DOUBLE PRECISION, DIMENSION(3), INTENT(IN) :: TAUVEC
      DOUBLE PRECISION, DIMENSION(3), INTENT(IN) :: DLTR
C      
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: DPVEC
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: CALG
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: TAUNEQ
C
      DOUBLE PRECISION, DIMENSION(6,6), INTENT(OUT) :: DDSDDE
C     
      DOUBLE PRECISION, DIMENSION(3,3,3,3) :: DDSDDETEMP
C
      DOUBLE PRECISION, PARAMETER :: eps = 1.0D-16
C
      integer :: K1
      integer :: K2
      integer :: K3
      integer :: K4
      DOUBLE PRECISION :: VAR1
      DOUBLE PRECISION :: VAR2
      DOUBLE PRECISION :: VAR3
C
      IF (ABS(DLTR(2)-DLTR(1)).LT.eps) THEN
      VAR1 = 0.0D0
      ELSE
      VAR1 = 1.0D0*(TAUVEC(2)*DLTR(1)-TAUVEC(1)*DLTR(2))
     1              /(DLTR(2)-DLTR(1))
      END IF
C
      IF (ABS(DLTR(3)-DLTR(1)).LT.eps) THEN
      VAR2 = 0.0D0
      ELSE
      VAR2 = 1.0D0*(TAUVEC(3)*DLTR(1)-TAUVEC(1)*DLTR(3))
     1              /(DLTR(3)-DLTR(1))
      END IF
C
      IF (ABS(DLTR(3)-DLTR(2)).LT.eps) THEN
      VAR3 = 0.0D0
      ELSE
      VAR3 = 1.0D0*(TAUVEC(3)*DLTR(2)-TAUVEC(2)*DLTR(3))
     1              /(DLTR(3)-DLTR(2))
      END IF
C
      DO K1 = 1, 3
        DO K2 = 1, 3
          DO K3 = 1, 3
            DO K4 = 1, 3
                DDSDDETEMP(K1,K2,K3,K4) =
     1   ( CALG(1,1)-2.0D0*TAUVEC(1) )
     1 *DPVEC(1,K1)*DPVEC(1,K2)*DPVEC(1,K3)*DPVEC(1,K4)
     1  +( CALG(2,2)-2.0D0*TAUVEC(2) )
     1 *DPVEC(2,K1)*DPVEC(2,K2)*DPVEC(2,K3)*DPVEC(2,K4)
     1  +( CALG(3,3)-2.0D0*TAUVEC(3) )
     1 *DPVEC(3,K1)*DPVEC(3,K2)*DPVEC(3,K3)*DPVEC(3,K4)
     1  +CALG(1,2)*DPVEC(1,K1)*DPVEC(1,K2)*DPVEC(2,K3)*DPVEC(2,K4)
     1  +CALG(1,3)*DPVEC(1,K1)*DPVEC(1,K2)*DPVEC(3,K3)*DPVEC(3,K4)
     1  +CALG(2,3)*DPVEC(2,K1)*DPVEC(2,K2)*DPVEC(3,K3)*DPVEC(3,K4)
     1  +CALG(2,1)*DPVEC(2,K1)*DPVEC(2,K2)*DPVEC(1,K3)*DPVEC(1,K4)
     1  +CALG(3,1)*DPVEC(3,K1)*DPVEC(3,K2)*DPVEC(1,K3)*DPVEC(1,K4)
     1  +CALG(3,2)*DPVEC(3,K1)*DPVEC(3,K2)*DPVEC(2,K3)*DPVEC(2,K4)
     1  +VAR1*DPVEC(1,K1)*DPVEC(2,K2)*DPVEC(1,K3)*DPVEC(2,K4)
     1  +VAR1*DPVEC(2,K1)*DPVEC(1,K2)*DPVEC(2,K3)*DPVEC(1,K4)
     1  +VAR1*DPVEC(1,K1)*DPVEC(2,K2)*DPVEC(2,K3)*DPVEC(1,K4)
     1  +VAR1*DPVEC(2,K1)*DPVEC(1,K2)*DPVEC(1,K3)*DPVEC(2,K4)
     1  +VAR2*DPVEC(1,K1)*DPVEC(3,K2)*DPVEC(1,K3)*DPVEC(3,K4)
     1  +VAR2*DPVEC(3,K1)*DPVEC(1,K2)*DPVEC(3,K3)*DPVEC(1,K4)
     1  +VAR2*DPVEC(1,K1)*DPVEC(3,K2)*DPVEC(3,K3)*DPVEC(1,K4)
     1  +VAR2*DPVEC(3,K1)*DPVEC(1,K2)*DPVEC(1,K3)*DPVEC(3,K4)
     1  +VAR3*DPVEC(2,K1)*DPVEC(3,K2)*DPVEC(2,K3)*DPVEC(3,K4)
     1  +VAR3*DPVEC(3,K1)*DPVEC(2,K2)*DPVEC(3,K3)*DPVEC(2,K4)
     1  +VAR3*DPVEC(2,K1)*DPVEC(3,K2)*DPVEC(3,K3)*DPVEC(2,K4)
     1  +VAR3*DPVEC(3,K1)*DPVEC(2,K2)*DPVEC(2,K3)*DPVEC(3,K4)
            END DO
          END DO
        END DO
      END DO
C
      DDSDDE(1,1)=DDSDDETEMP(1,1,1,1)/DET+2.0D0*TAUNEQ(1,1)/DET
      DDSDDE(2,1)=DDSDDETEMP(2,2,1,1)/DET
      DDSDDE(3,1)=DDSDDETEMP(3,3,1,1)/DET
      DDSDDE(4,1)=DDSDDETEMP(1,2,1,1)/DET+TAUNEQ(2,1)/DET
      DDSDDE(5,1)=DDSDDETEMP(1,3,1,1)/DET+TAUNEQ(3,1)/DET
      DDSDDE(6,1)=DDSDDETEMP(2,3,1,1)/DET
C
      DDSDDE(1,2)=DDSDDETEMP(1,1,2,2)/DET
      DDSDDE(2,2)=DDSDDETEMP(2,2,2,2)/DET+2.0D0*TAUNEQ(2,2)/DET
      DDSDDE(3,2)=DDSDDETEMP(3,3,2,2)/DET
      DDSDDE(4,2)=DDSDDETEMP(1,2,2,2)/DET+TAUNEQ(1,2)/DET
      DDSDDE(5,2)=DDSDDETEMP(1,3,2,2)/DET
      DDSDDE(6,2)=DDSDDETEMP(2,3,2,2)/DET+TAUNEQ(3,2)/DET
C
      DDSDDE(1,3)=DDSDDETEMP(1,1,3,3)/DET
      DDSDDE(2,3)=DDSDDETEMP(2,2,3,3)/DET
      DDSDDE(3,3)=DDSDDETEMP(3,3,3,3)/DET+2.0D0*TAUNEQ(3,3)/DET
      DDSDDE(4,3)=DDSDDETEMP(1,2,3,3)/DET
      DDSDDE(5,3)=DDSDDETEMP(1,3,3,3)/DET+TAUNEQ(1,3)/DET
      DDSDDE(6,3)=DDSDDETEMP(2,3,3,3)/DET+TAUNEQ(2,3)/DET
C
      DDSDDE(1,4)=DDSDDETEMP(1,1,1,2)/DET+TAUNEQ(1,2)/DET
      DDSDDE(2,4)=DDSDDETEMP(2,2,1,2)/DET+TAUNEQ(2,1)/DET
      DDSDDE(3,4)=DDSDDETEMP(3,3,1,2)/DET
      DDSDDE(4,4)=DDSDDETEMP(1,2,1,2)/DET
     1           +0.5D0*(TAUNEQ(1,1)+TAUNEQ(2,2))/DET
      DDSDDE(5,4)=DDSDDETEMP(1,3,1,2)/DET+0.5D0*TAUNEQ(3,2)/DET
      DDSDDE(6,4)=DDSDDETEMP(2,3,1,2)/DET+0.5D0*TAUNEQ(3,1)/DET
C
      DDSDDE(1,5)=DDSDDETEMP(1,1,1,3)/DET+TAUNEQ(1,3)/DET
      DDSDDE(2,5)=DDSDDETEMP(2,2,1,3)/DET
      DDSDDE(3,5)=DDSDDETEMP(3,3,1,3)/DET+TAUNEQ(3,1)/DET
      DDSDDE(4,5)=DDSDDETEMP(1,2,1,3)/DET+0.5D0*TAUNEQ(2,3)/DET
      DDSDDE(5,5)=DDSDDETEMP(1,3,1,3)/DET
     1           +0.5D0*(TAUNEQ(1,1)+TAUNEQ(3,3))/DET
      DDSDDE(6,5)=DDSDDETEMP(2,3,1,3)/DET+0.5D0*TAUNEQ(2,1)/DET
C
      DDSDDE(1,6)=DDSDDETEMP(1,1,2,3)/DET
      DDSDDE(2,6)=DDSDDETEMP(2,2,2,3)/DET+TAUNEQ(2,3)/DET
      DDSDDE(3,6)=DDSDDETEMP(3,3,2,3)/DET+TAUNEQ(3,2)/DET
      DDSDDE(4,6)=DDSDDETEMP(1,2,2,3)/DET+0.5D0*TAUNEQ(1,3)/DET
      DDSDDE(5,6)=DDSDDETEMP(1,3,2,3)/DET+0.5D0*TAUNEQ(1,2)/DET
      DDSDDE(6,6)=DDSDDETEMP(2,3,2,3)/DET
     1           +0.5D0*(TAUNEQ(2,2)+TAUNEQ(3,3))/DET
C
      RETURN
C
      END SUBROUTINE DNEQTANGENT