      SUBROUTINE UMAT(STRESS,STATEV,DDSDDE,SSE,SPD,SCD,
     1 RPL,DDSDDT,DRPLDE,DRPLDT,STRAN,DSTRAN,
     2 TIME,DTIME,TEMP,DTEMP,PREDEF,DPRED,MATERL,NDI,NSHR,NTENS,
     3 NSTATV,PROPS,NPROPS,COORDS,DROT,PNEWDT,CELENT,
     4 DFGRD0,DFGRD1,NOEL,NPT,KSLAY,KSPT,KSTEP,KINC)
C
      INCLUDE 'ABA_PARAM.INC'
C
      CHARACTER*8 MATERL
      DIMENSION STRESS(NTENS),STATEV(NSTATV),
     1 DDSDDE(NTENS,NTENS),DDSDDT(NTENS),DRPLDE(NTENS),
     2 STRAN(NTENS),DSTRAN(NTENS),DFGRD0(3,3),DFGRD1(3,3),
     3 TIME(2),PREDEF(1),DPRED(1),PROPS(NPROPS),COORDS(3),DROT(3,3)

C LOCAL ARRAYS
C ----------------------------------------------------------------
C EELAS - LOGARITHMIC ELASTIC STRAINS
C EELASP - PRINCIPAL ELASTIC STRAINS
C BBAR - DEVIATORIC RIGHT CAUCHY-GREEN TENSOR
C BBARP - PRINCIPAL VALUES OF BBAR
C BBARN - PRINCIPAL DIRECTION OF BBAR (AND EELAS)
C DISTGR - DEVIATORIC DEFORMATION GRADIENT (DISTORTION TENSOR)
C Ci - RIGHT CAUCHY GREEN STRAIN OF THE INELASTIC DEFORMATION GRADIENT
C Be - LEFT CAUCHY GREEN STRAIN OF THE ELASTIC DEFORMATION GRADIENT
C TAUNEQ - NEQ (VISCOUS) KIRCHHOFF STRESS
C TAUVEC, DLTR, DPVEC, CALG - VARIABLES TO CALULATE THE NEQ-TANGENT
C      (SEE SUBROUTINE DNEQTANGENT)
C DDSDDE1, DDSDDE2 - EQ-TANGENT AND NEQ-TANGENT MODULUS
C sigVISC - NEQ CAUCHY STRESS (VISCOUS STRESS)
C ----------------------------------------------------------------
C
      DIMENSION EELAS(6), EELASP(3), BBAR(6), BBARP(3), BBARN(3, 3),
     1 DISTGR(3,3), CiINV0(3,3), CiINV1(3,3), Be(3,3),
     2 TAUNEQ(3,3), TAUVEC(3), DLTR(3), DPVEC(3,3), CALG(3,3),
     3 DDSDDE2(6,6), DDSDDE1(NTENS,NTENS), BeINV(3,3), CiTEMP(3,3),
     4 Ci(3,3), sigVISC(6)
C
      PARAMETER(ZERO=0.D0, ONE=1.D0, TWO=2.D0, THREE=3.D0, FOUR=4.D0,
     1 FIVE=5.D0, SIX=6.D0, EIGHT=8.D0, TWELVE=12.D0)
C
C ----------------------------------------------------------------
C UMAT FOR COMPRESSIBLE HYPER-VISCOELASTIC MATERIAL
C CANNOT BE USED FOR PLANE STRESS
C ----------------------------------------------------------------
C PROPS(1) - C10
C PROPS(2) - C20 <0.1*C10
C PROPS(3) - C30 <0.01*C10
C PROPS(4) - D1 ~2/K
C PROPS(5) - D2 ~>10*D1
C PROPS(6) - D3 ~>100*D1
C PROPS(7) - VISCDEV ~TAU/(BETA*2C10)
C PROPS(8) - VISCVOL ~TAU/(BETA*2/D1)
C PROPS(9) - BETA ~1
C ----------------------------------------------------------------
C
C ELASTIC PROPERTIES
C
      C10=PROPS(1)
      C20=PROPS(2)
      C30=PROPS(3)
      D1=PROPS(4)
      D2=PROPS(5)
      D3=PROPS(6)
      VISCdev=PROPS(7)
      VISCvol=PROPS(8)
      BETA=PROPS(9)
      
C
C JACOBIAN AND DISTORTION TENSOR
C
      DET=DFGRD1(1, 1)*DFGRD1(2, 2)*DFGRD1(3, 3)
     1 -DFGRD1(1, 2)*DFGRD1(2, 1)*DFGRD1(3, 3)
      IF(NSHR.EQ.3) THEN
      DET=DET+DFGRD1(1, 2)*DFGRD1(2, 3)*DFGRD1(3, 1)
     1 +DFGRD1(1, 3)*DFGRD1(3, 2)*DFGRD1(2, 1)
     2 -DFGRD1(1, 3)*DFGRD1(3,1)*DFGRD1(2, 2)
     3 -DFGRD1(2, 3)*DFGRD1(3, 2)*DFGRD1(1, 1)
      END IF
      SCALE=DET**(-ONE/THREE)
      DO K1=1, 3
        DO K2=1, 3
            DISTGR(K2, K1)=SCALE*DFGRD1(K2, K1)
        END DO
      END DO
C CALCULATE DEVIATORIC LEFT CAUCHY-GREEN DEFORMATION TENSOR
C
      BBAR(1)=DISTGR(1, 1)**2+DISTGR(1, 2)**2+DISTGR(1, 3)**2
      BBAR(2)=DISTGR(2, 1)**2+DISTGR(2, 2)**2+DISTGR(2, 3)**2
      BBAR(3)=DISTGR(3, 3)**2+DISTGR(3, 1)**2+DISTGR(3, 2)**2
      BBAR(4)=DISTGR(1, 1)*DISTGR(2, 1)+DISTGR(1, 2)*DISTGR(2, 2)
     1 +DISTGR(1, 3)*DISTGR(2, 3)
      IF(NSHR.EQ.3) THEN
      BBAR(5)=DISTGR(1, 1)*DISTGR(3, 1)+DISTGR(1, 2)*DISTGR(3, 2)
     1 +DISTGR(1, 3)*DISTGR(3, 3)
      BBAR(6)=DISTGR(2, 1)*DISTGR(3, 1)+DISTGR(2, 2)*DISTGR(3, 2)
     1 +DISTGR(2, 3)*DISTGR(3, 3)
      END IF
C
C     DEFINING THE INITIAL STATE-VARIABLE INV(Ci)
C    
      IF (KSTEP.EQ.1 .AND. KINC.EQ.1) THEN
       STATEV(1)=ONE
       STATEV(2)=ONE
       STATEV(3)=ONE
       STATEV(4)=ZERO
       STATEV(5)=ZERO
       STATEV(6)=ZERO
      END IF
C      
      CiINV0(1,1)=STATEV(1)
      CiINV0(1,2)=STATEV(4)
      CiINV0(1,3)=STATEV(5)
      CiINV0(2,1)=STATEV(4)
      CiINV0(2,2)=STATEV(2)
      CiINV0(2,3)=STATEV(6)
      CiINV0(3,1)=STATEV(5)
      CiINV0(3,2)=STATEV(6)
      CiINV0(3,3)=STATEV(3)
C
C     CALCULATION OF ELASTIC LEFT CAUCHY-GREEN STRAIN AND 
C     KIRCHHOFF STRESS FOR THE NON-EQ (VISCOUS) DEFORMATION
C
      DTID=DTIME
        CALL BECALCMOD (C10, C20, C30, D1, D2, D3,BETA,NDI,NSHR,
     1            VISCdev, VISCvol, DTID,
     1            DFGRD1, CiINV0, Be, TAUNEQ, TAUVEC, DLTR, DPVEC, CALG)
    
C     MATERIAL TANGENT OF THE INELASTIC DEFORMATION
C
        CALL DNEQTANGENT (DET, TAUVEC, DLTR, DPVEC,
     1                        CALG, TAUNEQ, DDSDDE2)
C
C     CALCULATION OF RIGHT CAUCHY-GREEN STRAIN OF THE NON-EQ
C     DEFORMATION THE INVERSE TO UPDATE THE STATE-VARIABLES
C
      CALL M33INV(Be, BeINV, FLAG)
      CiTEMP=MATMUL(BeINV,DFGRD1)
      Ci=MATMUL(TRANSPOSE(DFGRD1),CiTEMP)
      CALL M33INV(Ci, CiINV1, FLAG)
C     
      STATEV(1)=CiINV1(1,1)
      STATEV(2)=CiINV1(2,2)
      STATEV(3)=CiINV1(3,3)
      STATEV(4)=CiINV1(1,2)
      STATEV(5)=CiINV1(1,3)
      STATEV(6)=CiINV1(2,3)
C
C     CALCUATION OF NON-EQ CAUCHY STRESS (VISCOUS STRESS)
C
      sigVISC(1)=TAUNEQ(1,1)/DET
      sigVISC(2)=TAUNEQ(2,2)/DET
      sigVISC(3)=TAUNEQ(3,3)/DET
      sigVISC(4)=TAUNEQ(1,2)/DET
      sigVISC(5)=TAUNEQ(1,3)/DET
      sigVISC(6)=TAUNEQ(2,3)/DET
C
C
C CALCULATE THE STRESS
C    
C     PARAMETERS
      TRBBAR=(BBAR(1)+BBAR(2)+BBAR(3))
      b1=TWO*C10+FOUR*C20*(TRBBAR-THREE)+SIX*C30*(TRBBAR-THREE)**2
      b2=EIGHT*C20+TWELVE*TWO*C30*(TRBBAR-THREE)
      a1=b1/DET
      a2=(TRBBAR/THREE*b2+TWO/THREE*b1)/DET
      a3=b2/DET
      a4=((TRBBAR/THREE)**2*b2+TWO/THREE*TRBBAR/THREE*b1)/DET
      PR=TWO/D1*(DET-ONE)+FOUR/D2*(DET-ONE)**3+SIX/D3*(DET-ONE)**5
      DP=TWO/D1*(TWO*DET-ONE)
     1  +FOUR/D2*((DET-ONE)**3+THREE*DET*(DET-ONE)**2)
     2  +SIX/D3*((DET-ONE)**5+FIVE*DET*(DET-ONE)**4)
C     
      DO K1=1,NDI
        STRESS(K1)=a1*(BBAR(K1)-TRBBAR/THREE)+PR+sigVISC(K1)
      END DO
      DO K1=NDI+1,NDI+NSHR
      STRESS(K1)=a1*BBAR(K1)+sigVISC(K1)
      END DO
C CALCULATE THE STIFFNESS    
C
      DDSDDE1(1, 1)= a4+TWO*(a1-a2)*BBAR(1)+a3*BBAR(1)**2+DP
      DDSDDE1(2, 2)= a4+TWO*(a1-a2)*BBAR(2)+a3*BBAR(2)**2+DP
      DDSDDE1(3, 3)= a4+TWO*(a1-a2)*BBAR(3)+a3*BBAR(3)**2+DP
      DDSDDE1(1, 2)= a4-a2*(BBAR(1)+BBAR(2))+a3*BBAR(1)*BBAR(2)+DP
      DDSDDE1(1, 3)= a4-a2*(BBAR(1)+BBAR(3))+a3*BBAR(1)*BBAR(3)+DP
      DDSDDE1(2, 3)= a4-a2*(BBAR(2)+BBAR(3))+a3*BBAR(2)*BBAR(3)+DP
      DDSDDE1(1, 4)= BBAR(4)*a1 - BBAR(4)*a2 + BBAR(1)*BBAR(4)*a3
      DDSDDE1(2, 4)= BBAR(4)*a1 - BBAR(4)*a2 + BBAR(2)*BBAR(4)*a3
      DDSDDE1(3, 4)= BBAR(3)*BBAR(4)*a3 - BBAR(4)*a2
      DDSDDE1(4, 4)= a3*BBAR(4)**2 + a1*(BBAR(1)/TWO + BBAR(2)/TWO)
      IF(NSHR.EQ.3) THEN
      DDSDDE1(1, 5)= BBAR(5)*a1 - BBAR(5)*a2 + BBAR(1)*BBAR(5)*a3
      DDSDDE1(2, 5)= BBAR(2)*BBAR(5)*a3 - BBAR(5)*a2
      DDSDDE1(3, 5)= BBAR(5)*a1 - BBAR(5)*a2 + BBAR(3)*BBAR(5)*a3
      DDSDDE1(1, 6)= BBAR(1)*BBAR(6)*a3 - BBAR(6)*a2
      DDSDDE1(2, 6)= BBAR(6)*a1 - BBAR(6)*a2 + BBAR(2)*BBAR(6)*a3
      DDSDDE1(3, 6)= BBAR(6)*a1 - BBAR(6)*a2 + BBAR(3)*BBAR(6)*a3
      DDSDDE1(5, 5)= a3*BBAR(5)**2 + a1*(BBAR(1)/TWO + BBAR(3)/TWO)
      DDSDDE1(6, 6)= a3*BBAR(6)**2 + a1*(BBAR(2)/TWO + BBAR(3)/TWO)
      DDSDDE1(4,5)= (BBAR(6)*a1)/TWO + BBAR(4)*BBAR(5)*a3
      DDSDDE1(4,6)= (BBAR(5)*a1)/TWO + BBAR(4)*BBAR(6)*a3
      DDSDDE1(5,6)= (BBAR(4)*a1)/TWO + BBAR(5)*BBAR(6)*a3
      END IF
      DO K1=1, NTENS
      DO K2=1, K1-1
      DDSDDE1(K1, K2)=DDSDDE1(K2, K1)
      END DO
      END DO
      IF(NSHR.EQ.3) THEN
        DO K1=1, 6
            DO K2=1, 6
                DDSDDE(K1,K2)=DDSDDE1(K1,K2)+DDSDDE2(K1,K2)
            END DO
            END DO
      ELSE
        DO K1=1, 4
            DO K2=1, 4
      DDSDDE(K1,K2)=DDSDDE1(K1,K2)+DDSDDE2(K1,K2)
        END DO
            END DO
      END IF
C
C STORE ELASTIC STRAINS IN STATE VARIABLE ARRAY
C
      STATEV(7)=DET
      STATEV(8)=STRESS(2)
      STATEV(9)=STRESS(2)-sigVISC(2)
      STATEV(10)=sigVISC(2)
      STATEV(11)=DFGRD1(2,2)
      STATEV(12)=DFGRD1(1,1)
      STATEV(13)=Be(2,2)
      STATEV(14)=Be(1,1)
      STATEV(15)=-LOG(DFGRD1(1,1))/LOG(DFGRD1(2,2))
C
      RETURN
      END
      
      
!***********************************************************************************************************************************
!  BECALCMOD -  CALCULATE ELASTIC LEFT CAUCHY STRAIN AND NEQ-KIRCHOFF STRESS
!  
!  INPUT VARIABLES:
!  DFGRD = DEFORMATION GRADIENT OF STEP K
!  CiINV = INVERSE INELASTIC (VISCOUS) RIGHT CAUCHY-GREEN STRAIN TENSOR AT STEP K-1
!  TAU   = KIRCHOFF STRESS AT STEP K+1
!  MATERIALPARAMETERS = DMU, DK, BETA, VISCdev, VISCvol
!  
!  OUTPUT VARIABLES:
!  Be = ELASTIC LEFT CAUCHY STRAIN
!  TAUNEQ = NEQ-KIRCHOFF STRESS
!***********************************************************************************************************************************
C
      SUBROUTINE BECALCMOD (C1, C2, C3, D1, D2, D3, BETA, NDI,NSHR,
     1                   VISCdev, VISCvol, DTIME,
     1                   DFGRD, CiINV, 
     1                   Be, TAUNEQ, TAUVEC, DLTR, DPVEC, CALG)

      IMPLICIT NONE

      DOUBLE PRECISION, INTENT(IN) :: C1
      DOUBLE PRECISION, INTENT(IN) :: C2
      DOUBLE PRECISION, INTENT(IN) :: C3
      DOUBLE PRECISION, INTENT(IN) :: D1
      DOUBLE PRECISION, INTENT(IN) :: D2
      DOUBLE PRECISION, INTENT(IN) :: D3
      DOUBLE PRECISION, INTENT(IN) :: BETA
      DOUBLE PRECISION, INTENT(IN) :: NDI
      DOUBLE PRECISION, INTENT(IN) :: NSHR
      DOUBLE PRECISION, INTENT(IN) :: VISCdev
      DOUBLE PRECISION, INTENT(IN) :: VISCvol
      DOUBLE PRECISION, INTENT(IN) :: DTIME
C      
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: DFGRD
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: CiINV
C
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: Be
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: TAUNEQ
      DOUBLE PRECISION, DIMENSION(3), INTENT(OUT) :: TAUVEC
      DOUBLE PRECISION, DIMENSION(3), INTENT(OUT) :: DLTR
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: DPVEC
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: CALG
C
      DOUBLE PRECISION, PARAMETER :: TOL = 1.0D-12
C     
      DOUBLE PRECISION, DIMENSION(3,3) :: BeTEMP
      DOUBLE PRECISION, DIMENSION(3,3) :: BeTRIAL
      DOUBLE PRECISION, DIMENSION(6) :: BeTR_
C
      DOUBLE PRECISION, DIMENSION(3) :: PV
      DOUBLE PRECISION, DIMENSION(3,3) :: PD
C
      DOUBLE PRECISION :: V1
      DOUBLE PRECISION :: V2
      DOUBLE PRECISION :: BI1
      DOUBLE PRECISION :: DJE0
      DOUBLE PRECISION :: TRACE
      DOUBLE PRECISION :: B1
      DOUBLE PRECISION :: B2
      DOUBLE PRECISION :: P
      DOUBLE PRECISION :: DP
      DOUBLE PRECISION :: DEV1
      DOUBLE PRECISION :: DEV2
      DOUBLE PRECISION :: DEV3
      DOUBLE PRECISION, DIMENSION(3) :: Q0
      DOUBLE PRECISION, DIMENSION(3) :: Q1
      DOUBLE PRECISION, DIMENSION(3) :: Q0BAR
      DOUBLE PRECISION, DIMENSION(3) :: E0
      DOUBLE PRECISION, DIMENSION(3) :: E1
      DOUBLE PRECISION, DIMENSION(3) :: DE
      DOUBLE PRECISION, DIMENSION(3) :: ETRIAL
C
      DOUBLE PRECISION, DIMENSION(3,3) :: DKMAT
      DOUBLE PRECISION, DIMENSION(3,3) :: DKINV
      DOUBLE PRECISION, DIMENSION(3) :: VR
      DOUBLE PRECISION, DIMENSION(3,3) :: DTDEPS
C
      integer :: K1
      integer :: K2
      integer :: K3
      DOUBLE PRECISION :: DnormDE
      LOGICAL :: FLAG
C
C
C
C      
      BeTEMP = MATMUL(CiINV,TRANSPOSE(DFGRD))
      BeTRIAL = MATMUL(DFGRD,BeTEMP)
      BeTR_(1) = BeTRIAL(1,1)
      BeTR_(2) = BeTRIAL(2,2)
      BeTR_(3) = BeTRIAL(3,3)
      BeTR_(4) = BeTRIAL(1,2)+BeTRIAL(2,1)
      BeTR_(5) = BeTRIAL(1,3)+BeTRIAL(3,1)
      BeTR_(6) = BeTRIAL(2,3)+BeTRIAL(3,2)
C
      CALL SPRIND(BeTR_,PV,PD,2,NDI,NSHR)
C     
      V1=DTIME*BETA/(2.D0*VISCdev)
      V2=DTIME*BETA/(3.D0*VISCvol)
C     
      Q0(1)=PV(1)
      Q0(2)=PV(2)
      Q0(3)=PV(3)
      ETRIAL(1)=1.D0/2.D0*LOG(PV(1))
      ETRIAL(2)=1.D0/2.D0*LOG(PV(2))
      ETRIAL(3)=1.D0/2.D0*LOG(PV(3))
C
      do K3=1,175,1
        DJE0=(Q0(1)*Q0(2)*Q0(3))**(1.0D0/2.0D0)
        E0(1)=1.D0/2.D0*LOG(Q0(1))
        E0(2)=1.D0/2.D0*LOG(Q0(2))
        E0(3)=1.D0/2.D0*LOG(Q0(3))
        Q0BAR(1)=DJE0**(-2.D0/3.D0)*Q0(1)
        Q0BAR(2)=DJE0**(-2.D0/3.D0)*Q0(2)
        Q0BAR(3)=DJE0**(-2.D0/3.D0)*Q0(3)
        TRACE=Q0BAR(1)+Q0BAR(2)+Q0BAR(3)
        DEV1=2.D0/3.D0*Q0BAR(1)-1.D0/3.D0*(Q0BAR(2)+Q0BAR(3))
        DEV2=2.D0/3.D0*Q0BAR(2)-1.D0/3.D0*(Q0BAR(1)+Q0BAR(3))
        DEV3=2.D0/3.D0*Q0BAR(3)-1.D0/3.D0*(Q0BAR(1)+Q0BAR(2))
        B1=2.D0*C1+4.D0*C2*(TRACE-3.D0)+6.D0*(TRACE-3.D0)**2.D0
        B2=8.D0*C2+24.D0*C3*(TRACE-3.D0)
        P=2.D0/D1*(DJE0-1.D0)+4.D0/D2*(DJE0-1.D0)**3.D0
     1   +6.D0/D3*(DJE0-1.D0)**5.D0
        DP=2.D0/D1*(2.D0*DJE0-1.D0)
     1    +4.D0/D2*( (DJE0-1.D0)**3.D0+3.D0*DJE0*(DJE0-1.D0)**2.D0 )
     2    +6.D0/D3*( (DJE0-1.D0)**5.D0+5.D0*DJE0*(DJE0-1.D0)**4.D0 )
C        
        DKMAT(1,1)=V1*(B2*DEV1*DEV1
     1 +B1*2.D0/3.D0*(4.D0/3.D0*Q0BAR(1)+1.D0/3.D0*(Q0BAR(2)+Q0BAR(3))))
     2 +V2*DP*DJE0+1.D0
        DKMAT(1,2)=V1*(B2*DEV1*DEV2
     1 +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(3)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(2))))
     2 +V2*DP*DJE0
        DKMAT(1,3)=V1*(B2*DEV1*DEV3
     1 +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(2)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(3))))
     2 +V2*DP*DJE0
        DKMAT(2,1)=V1*(B2*DEV2*DEV1
     1 +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(3)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(2))))
     2 +V2*DP*DJE0
        DKMAT(2,2)=V1*(B2*DEV2*DEV2
     1 +B1*2.D0/3.D0*(4.D0/3.D0*Q0BAR(2)+1.D0/3.D0*(Q0BAR(1)+Q0BAR(3))))
     2 +V2*DP*DJE0+1.D0
        DKMAT(2,3)=V1*(B2*DEV2*DEV3
     1 +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(1)-2.D0/3.D0*(Q0BAR(2)+Q0BAR(3))))
     2 +V2*DP*DJE0
        DKMAT(3,1)=V1*(B2*DEV3*DEV1
     1 +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(2)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(3))))
     2 +V2*DP*DJE0
        DKMAT(3,2)=V1*(B2*DEV3*DEV2
     1 +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(1)-2.D0/3.D0*(Q0BAR(2)+Q0BAR(3))))
     2 +V2*DP*DJE0
        DKMAT(3,3)=V1*(B2*DEV3*DEV3
     1 +B1*2.D0/3.D0*(4.D0/3.D0*Q0BAR(3)+1.D0/3.D0*(Q0BAR(2)+Q0BAR(1))))
     2 +V2*DP*DJE0+1.D0
C       
        VR(1)=V1*B1*DEV1+V2*P*DJE0+E0(1)-ETRIAL(1)
        VR(2)=V1*B1*DEV2+V2*P*DJE0+E0(2)-ETRIAL(2)
        VR(3)=V1*B1*DEV3+V2*P*DJE0+E0(3)-ETRIAL(3)
C       
        CALL M33INV(DKMAT, DKINV, FLAG)
        DE(1)=-(DKINV(1,1)*VR(1)+DKINV(1,2)*VR(2)+DKINV(1,3)*VR(3))
        DE(2)=-(DKINV(2,1)*VR(1)+DKINV(2,2)*VR(2)+DKINV(2,3)*VR(3))
        DE(3)=-(DKINV(3,1)*VR(1)+DKINV(3,2)*VR(2)+DKINV(3,3)*VR(3))
C       
        E1(1)=E0(1)+DE(1)
        E1(2)=E0(2)+DE(2)
        E1(3)=E0(3)+DE(3)
C
        DnormDE=(DE(1)**2.D0+DE(2)**2.D0+DE(3)**2.D0)**(1.D0/2.D0)
        if (abs(DnormDE).LT.TOL) then 
            DO K1 = 1, 3
                DO K2 = 1, 3
                    Be(K2,K1)=Q0(1)*PD(1,K2)*PD(1,K1)
     1                       +Q0(2)*PD(2,K2)*PD(2,K1)
     1                       +Q0(3)*PD(3,K2)*PD(3,K1)
                END DO
            END DO
            BI1=Q0(1)+Q0(2)+Q0(3)
            TAUNEQ(1,1)=BETA*P*DJE0
     1                 +BETA*B1*DJE0**(-2.0D0/3.0D0)
     1                 *(Be(1,1)-1.0D0/3.0D0*BI1)
            TAUNEQ(2,2)=BETA*P*DJE0
     1                 +BETA*B1*DJE0**(-2.0D0/3.0D0)
     1                 *(Be(2,2)-1.0D0/3.0D0*BI1)
            TAUNEQ(3,3)=BETA*P*DJE0
     1                 +BETA*B1*DJE0**(-2.0D0/3.0D0)
     1                 *(Be(3,3)-1.0D0/3.0D0*BI1)
            TAUNEQ(1,2)=BETA*B1*DJE0**(-2.0D0/3.0D0)*Be(1,2)
            TAUNEQ(1,3)=BETA*B1*DJE0**(-2.0D0/3.0D0)*Be(1,3)
            TAUNEQ(2,3)=BETA*B1*DJE0**(-2.0D0/3.0D0)*Be(2,3)
            TAUNEQ(2,1)=BETA*B1*DJE0**(-2.0D0/3.0D0)*Be(2,1)
            TAUNEQ(3,1)=BETA*B1*DJE0**(-2.0D0/3.0D0)*Be(3,1)
            TAUNEQ(3,2)=BETA*B1*DJE0**(-2.0D0/3.0D0)*Be(3,2)
            TAUVEC(1)=BETA*P*DJE0
     1                 +BETA*B1*DJE0**(-2.0D0/3.0D0)
     1                 *(Q0(1)-1.0D0/3.0D0*BI1)
            TAUVEC(2)=BETA*P*DJE0
     1                 +BETA*B1*DJE0**(-2.0D0/3.0D0)
     1                 *(Q0(2)-1.0D0/3.0D0*BI1)
            TAUVEC(3)=BETA*P*DJE0
     1                 +BETA*B1*DJE0**(-2.0D0/3.0D0)
     1                 *(Q0(3)-1.0D0/3.0D0*BI1)
            DLTR(1)=EXP(2.0D0*ETRIAL(1))
            DLTR(2)=EXP(2.0D0*ETRIAL(2))
            DLTR(3)=EXP(2.0D0*ETRIAL(3))
            DO K1 = 1, 3
                DO K2 = 1, 3
                    DPVEC(K2,K1)=PD(K2,K1)
                END DO
            END DO
        DTDEPS(1,1)=BETA*(  B2*DEV1*DEV1
     1  +B1*2.D0/3.D0*(4.D0/3.D0*Q0BAR(1)+1.D0/3.D0*(Q0BAR(2)+Q0BAR(3)))
     2  +DP*DJE0 )
        DTDEPS(1,2)=BETA*(  B2*DEV1*DEV2
     1  +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(3)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(2)))
     2  +DP*DJE0 )
        DTDEPS(1,3)=BETA*(  B2*DEV1*DEV3
     1  +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(2)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(3)))
     2  +DP*DJE0 )
        DTDEPS(2,1)=BETA*(  B2*DEV2*DEV1
     1  +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(3)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(2)))
     2  +DP*DJE0 )
        DTDEPS(2,2)=BETA*(  B2*DEV2*DEV2
     1  +B1*2.D0/3.D0*(4.D0/3.D0*Q0BAR(2)+1.D0/3.D0*(Q0BAR(1)+Q0BAR(3)))
     2  +DP*DJE0 )
        DTDEPS(2,3)=BETA*(  B2*DEV2*DEV3
     1  +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(1)-2.D0/3.D0*(Q0BAR(2)+Q0BAR(3)))
     2  +DP*DJE0 )
        DTDEPS(3,1)=BETA*(  B2*DEV3*DEV1
     1  +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(2)-2.D0/3.D0*(Q0BAR(1)+Q0BAR(3)))
     2  +DP*DJE0 )
        DTDEPS(3,2)=BETA*(  B2*DEV3*DEV2
     1  +B1*2.D0/3.D0*(1.D0/3.D0*Q0BAR(1)-2.D0/3.D0*(Q0BAR(2)+Q0BAR(3)))
     2  +DP*DJE0 )
        DTDEPS(3,3)=BETA*(  B2*DEV3*DEV3
     1  +B1*2.D0/3.D0*(4.D0/3.D0*Q0BAR(3)+1.D0/3.D0*(Q0BAR(2)+Q0BAR(1)))
     2  +DP*DJE0 )
            CALG = MATMUL(DTDEPS,DKINV)
            return
        end if 
C     
      Q1(1)=EXP(2.D0*E1(1))
      Q1(2)=EXP(2.D0*E1(2))
      Q1(3)=EXP(2.D0*E1(3))
      Q0(1)=Q1(1)
      Q0(2)=Q1(2)
      Q0(3)=Q1(3)
C        
      END DO
C
      RETURN
C
      END SUBROUTINE BECALCMOD
      
!***********************************************************************************************************************************
!  M33INV  -  Compute the inverse of a 3x3 matrix.
!
!  A       = input 3x3 matrix to be inverted
!  AINV    = output 3x3 inverse of matrix A
!  OK_FLAG = (output) .TRUE. if the input matrix could be inverted, and .FALSE. if the input matrix is singular.
!  Code written by:   David G. Simpson  -   NASA Goddard Space Flight Center
!***********************************************************************************************************************************

      SUBROUTINE M33INV (A, AINV, OK_FLAG)

      IMPLICIT NONE

      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN)  :: A
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: AINV
      LOGICAL, INTENT(OUT) :: OK_FLAG

      DOUBLE PRECISION, PARAMETER :: EPS = 1.0D-16
      DOUBLE PRECISION :: DET
      DOUBLE PRECISION, DIMENSION(3,3) :: COFACTOR


      DET =   A(1,1)*A(2,2)*A(3,3)  
     1       - A(1,1)*A(2,3)*A(3,2)  
     1       - A(1,2)*A(2,1)*A(3,3)  
     1       + A(1,2)*A(2,3)*A(3,1)  
     1       + A(1,3)*A(2,1)*A(3,2)  
     1       - A(1,3)*A(2,2)*A(3,1)

      IF (ABS(DET) .LE. EPS) THEN
         AINV = 0.0D0
         OK_FLAG = .FALSE.
         RETURN
      END IF

      COFACTOR(1,1) = +(A(2,2)*A(3,3)-A(2,3)*A(3,2))
      COFACTOR(1,2) = -(A(2,1)*A(3,3)-A(2,3)*A(3,1))
      COFACTOR(1,3) = +(A(2,1)*A(3,2)-A(2,2)*A(3,1))
      COFACTOR(2,1) = -(A(1,2)*A(3,3)-A(1,3)*A(3,2))
      COFACTOR(2,2) = +(A(1,1)*A(3,3)-A(1,3)*A(3,1))
      COFACTOR(2,3) = -(A(1,1)*A(3,2)-A(1,2)*A(3,1))
      COFACTOR(3,1) = +(A(1,2)*A(2,3)-A(1,3)*A(2,2))
      COFACTOR(3,2) = -(A(1,1)*A(2,3)-A(1,3)*A(2,1))
      COFACTOR(3,3) = +(A(1,1)*A(2,2)-A(1,2)*A(2,1))

      AINV = TRANSPOSE(COFACTOR) / DET

      OK_FLAG = .TRUE.

      RETURN

      END SUBROUTINE M33INV
      

      
      
!***********************************************************************************************************************************
!  DNEQTANGENT -  CALCULATE NEQ TANGENT
!  
!  INPUT VARIABLES:
!  DET      =   DETERMINANT OF DEFORMATION GRADIENT
!  TAUVEC   =   VECTOR CONTAINING THE PRINCIPAL KIRCHOFF-STRESSES
!  DLTR     =   VECTOR CONTAINING THE SQUARED PRINCIPAL TRIAL-STRAINS
!  DPVEC    =   MATRIX CONTAINING THE THREE PRINCIPAL DIRECTIONS OF Be
!  CALG     =   ALGORITMIC MODULUS (3,3)
!  TAUNEQ   =   INELASTIC KIRCHOFF STRESS (3,3)
!  
!  OUTPUT VARIABLES:
!  DDSDDE = NEQ MATERIAL-TANGENT (6,6)
!***********************************************************************************************************************************

      SUBROUTINE DNEQTANGENT (DET, TAUVEC, DLTR, DPVEC,
     1                        CALG, TAUNEQ, DDSDDE)
C
      IMPLICIT NONE
C
      DOUBLE PRECISION, INTENT(IN) :: DET
      DOUBLE PRECISION, DIMENSION(3), INTENT(IN) :: TAUVEC
      DOUBLE PRECISION, DIMENSION(3), INTENT(IN) :: DLTR
C      
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: DPVEC
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: CALG
      DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN) :: TAUNEQ
C
      DOUBLE PRECISION, DIMENSION(6,6), INTENT(OUT) :: DDSDDE
C     
      DOUBLE PRECISION, DIMENSION(3,3,3,3) :: DDSDDETEMP
C
      DOUBLE PRECISION, PARAMETER :: eps = 1.0D-16
C
      integer :: K1
      integer :: K2
      integer :: K3
      integer :: K4
      DOUBLE PRECISION :: VAR1
      DOUBLE PRECISION :: VAR2
      DOUBLE PRECISION :: VAR3
C
      IF (ABS(DLTR(2)-DLTR(1)).LT.eps) THEN
      VAR1 = 0.0D0
      ELSE
      VAR1 = 1.0D0*(TAUVEC(2)*DLTR(1)-TAUVEC(1)*DLTR(2))
     1              /(DLTR(2)-DLTR(1))
      END IF
C
      IF (ABS(DLTR(3)-DLTR(1)).LT.eps) THEN
      VAR2 = 0.0D0
      ELSE
      VAR2 = 1.0D0*(TAUVEC(3)*DLTR(1)-TAUVEC(1)*DLTR(3))
     1              /(DLTR(3)-DLTR(1))
      END IF
C
      IF (ABS(DLTR(3)-DLTR(2)).LT.eps) THEN
      VAR3 = 0.0D0
      ELSE
      VAR3 = 1.0D0*(TAUVEC(3)*DLTR(2)-TAUVEC(2)*DLTR(3))
     1              /(DLTR(3)-DLTR(2))
      END IF
C
      DO K1 = 1, 3
        DO K2 = 1, 3
          DO K3 = 1, 3
            DO K4 = 1, 3
                DDSDDETEMP(K1,K2,K3,K4) =
     1   ( CALG(1,1)-2.0D0*TAUVEC(1) )
     1 *DPVEC(1,K1)*DPVEC(1,K2)*DPVEC(1,K3)*DPVEC(1,K4)
     1  +( CALG(2,2)-2.0D0*TAUVEC(2) )
     1 *DPVEC(2,K1)*DPVEC(2,K2)*DPVEC(2,K3)*DPVEC(2,K4)
     1  +( CALG(3,3)-2.0D0*TAUVEC(3) )
     1 *DPVEC(3,K1)*DPVEC(3,K2)*DPVEC(3,K3)*DPVEC(3,K4)
     1  +CALG(1,2)*DPVEC(1,K1)*DPVEC(1,K2)*DPVEC(2,K3)*DPVEC(2,K4)
     1  +CALG(1,3)*DPVEC(1,K1)*DPVEC(1,K2)*DPVEC(3,K3)*DPVEC(3,K4)
     1  +CALG(2,3)*DPVEC(2,K1)*DPVEC(2,K2)*DPVEC(3,K3)*DPVEC(3,K4)
     1  +CALG(2,1)*DPVEC(2,K1)*DPVEC(2,K2)*DPVEC(1,K3)*DPVEC(1,K4)
     1  +CALG(3,1)*DPVEC(3,K1)*DPVEC(3,K2)*DPVEC(1,K3)*DPVEC(1,K4)
     1  +CALG(3,2)*DPVEC(3,K1)*DPVEC(3,K2)*DPVEC(2,K3)*DPVEC(2,K4)
     1  +VAR1*DPVEC(1,K1)*DPVEC(2,K2)*DPVEC(1,K3)*DPVEC(2,K4)
     1  +VAR1*DPVEC(2,K1)*DPVEC(1,K2)*DPVEC(2,K3)*DPVEC(1,K4)
     1  +VAR1*DPVEC(1,K1)*DPVEC(2,K2)*DPVEC(2,K3)*DPVEC(1,K4)
     1  +VAR1*DPVEC(2,K1)*DPVEC(1,K2)*DPVEC(1,K3)*DPVEC(2,K4)
     1  +VAR2*DPVEC(1,K1)*DPVEC(3,K2)*DPVEC(1,K3)*DPVEC(3,K4)
     1  +VAR2*DPVEC(3,K1)*DPVEC(1,K2)*DPVEC(3,K3)*DPVEC(1,K4)
     1  +VAR2*DPVEC(1,K1)*DPVEC(3,K2)*DPVEC(3,K3)*DPVEC(1,K4)
     1  +VAR2*DPVEC(3,K1)*DPVEC(1,K2)*DPVEC(1,K3)*DPVEC(3,K4)
     1  +VAR3*DPVEC(2,K1)*DPVEC(3,K2)*DPVEC(2,K3)*DPVEC(3,K4)
     1  +VAR3*DPVEC(3,K1)*DPVEC(2,K2)*DPVEC(3,K3)*DPVEC(2,K4)
     1  +VAR3*DPVEC(2,K1)*DPVEC(3,K2)*DPVEC(3,K3)*DPVEC(2,K4)
     1  +VAR3*DPVEC(3,K1)*DPVEC(2,K2)*DPVEC(2,K3)*DPVEC(3,K4)
            END DO
          END DO
        END DO
      END DO
C
      DDSDDE(1,1)=DDSDDETEMP(1,1,1,1)/DET+2.0D0*TAUNEQ(1,1)/DET
      DDSDDE(2,1)=DDSDDETEMP(2,2,1,1)/DET
      DDSDDE(3,1)=DDSDDETEMP(3,3,1,1)/DET
      DDSDDE(4,1)=DDSDDETEMP(1,2,1,1)/DET+TAUNEQ(2,1)/DET
      DDSDDE(5,1)=DDSDDETEMP(1,3,1,1)/DET+TAUNEQ(3,1)/DET
      DDSDDE(6,1)=DDSDDETEMP(2,3,1,1)/DET
C
      DDSDDE(1,2)=DDSDDETEMP(1,1,2,2)/DET
      DDSDDE(2,2)=DDSDDETEMP(2,2,2,2)/DET+2.0D0*TAUNEQ(2,2)/DET
      DDSDDE(3,2)=DDSDDETEMP(3,3,2,2)/DET
      DDSDDE(4,2)=DDSDDETEMP(1,2,2,2)/DET+TAUNEQ(1,2)/DET
      DDSDDE(5,2)=DDSDDETEMP(1,3,2,2)/DET
      DDSDDE(6,2)=DDSDDETEMP(2,3,2,2)/DET+TAUNEQ(3,2)/DET
C
      DDSDDE(1,3)=DDSDDETEMP(1,1,3,3)/DET
      DDSDDE(2,3)=DDSDDETEMP(2,2,3,3)/DET
      DDSDDE(3,3)=DDSDDETEMP(3,3,3,3)/DET+2.0D0*TAUNEQ(3,3)/DET
      DDSDDE(4,3)=DDSDDETEMP(1,2,3,3)/DET
      DDSDDE(5,3)=DDSDDETEMP(1,3,3,3)/DET+TAUNEQ(1,3)/DET
      DDSDDE(6,3)=DDSDDETEMP(2,3,3,3)/DET+TAUNEQ(2,3)/DET
C
      DDSDDE(1,4)=DDSDDETEMP(1,1,1,2)/DET+TAUNEQ(1,2)/DET
      DDSDDE(2,4)=DDSDDETEMP(2,2,1,2)/DET+TAUNEQ(2,1)/DET
      DDSDDE(3,4)=DDSDDETEMP(3,3,1,2)/DET
      DDSDDE(4,4)=DDSDDETEMP(1,2,1,2)/DET
     1           +0.5D0*(TAUNEQ(1,1)+TAUNEQ(2,2))/DET
      DDSDDE(5,4)=DDSDDETEMP(1,3,1,2)/DET+0.5D0*TAUNEQ(3,2)/DET
      DDSDDE(6,4)=DDSDDETEMP(2,3,1,2)/DET+0.5D0*TAUNEQ(3,1)/DET
C
      DDSDDE(1,5)=DDSDDETEMP(1,1,1,3)/DET+TAUNEQ(1,3)/DET
      DDSDDE(2,5)=DDSDDETEMP(2,2,1,3)/DET
      DDSDDE(3,5)=DDSDDETEMP(3,3,1,3)/DET+TAUNEQ(3,1)/DET
      DDSDDE(4,5)=DDSDDETEMP(1,2,1,3)/DET+0.5D0*TAUNEQ(2,3)/DET
      DDSDDE(5,5)=DDSDDETEMP(1,3,1,3)/DET
     1           +0.5D0*(TAUNEQ(1,1)+TAUNEQ(3,3))/DET
      DDSDDE(6,5)=DDSDDETEMP(2,3,1,3)/DET+0.5D0*TAUNEQ(2,1)/DET
C
      DDSDDE(1,6)=DDSDDETEMP(1,1,2,3)/DET
      DDSDDE(2,6)=DDSDDETEMP(2,2,2,3)/DET+TAUNEQ(2,3)/DET
      DDSDDE(3,6)=DDSDDETEMP(3,3,2,3)/DET+TAUNEQ(3,2)/DET
      DDSDDE(4,6)=DDSDDETEMP(1,2,2,3)/DET+0.5D0*TAUNEQ(1,3)/DET
      DDSDDE(5,6)=DDSDDETEMP(1,3,2,3)/DET+0.5D0*TAUNEQ(1,2)/DET
      DDSDDE(6,6)=DDSDDETEMP(2,3,2,3)/DET
     1           +0.5D0*(TAUNEQ(2,2)+TAUNEQ(3,3))/DET
C
      RETURN
C
      END SUBROUTINE DNEQTANGENT